from dataclasses import dataclass


@dataclass
class FtpClientInitArgs:
    host: str
    user: str
    password: str
    client_name: str = "aioftp_client"
    port: int = 21


@dataclass
class ListFilesArgs:
    path: str


@dataclass
class ReadFileArgs:
    file_path: str



from pydantic import BaseModel, Field


class AsyncHttpClientInitArgs(BaseModel):
    access_token: str
    base_url: str
    client_name: str = "aiohttp_client"


class GetArgs(BaseModel):
    endpoint: str
    params: dict | None = Field(default=None)


class UploadFileArgs(BaseModel):
    endpoint: str
    file_bytes: bytes



from typing import Dict

from .controllers.update_thies_data import UpdateThiesDataController
from .controllers.types.update_thies_data_types import UpdateThiesDataControllerInput

class EpiiAPI:
    def update_thies_data(
        self,
        ftp_port: int,
        ftp_host: str,
        ftp_password: str,
        ftp_user: str,
    ) -> Dict[str, any]:
        controller = UpdateThiesDataController(
            UpdateThiesDataControllerInput(
                ftp_port=ftp_port,
                ftp_host=ftp_host,
                ftp_password=ftp_password,
                ftp_user=ftp_user,
            )
        )
        response = controller.execute()
        return response.__dict__



class EmptyDataError(Exception):
    def __str__(self):
        return "The data provided is empty."


class HttpClientError(Exception):
    def __str__(self):
        return "Http Client initialization fails."


class FtpClientError(Exception):
    def __str__(self):
        return "Ftp Client initialization fails."



PATH_AVG_FILES = "ftp/thies/BINFILES/ARCH_AV1"
PATH_EXT_FILES = "ftp/thies/BINFILES/ARCH_EX1"
DRIVE_ID = "b!Row14jaFrU-1q8qzrvj3OmPPTYWXizFEpJmI-wsfH5pXxA0qQwgQS50m2xvPCZem"
FILE_TYPES = ["AVG", "EXT"]



from abc import ABC, abstractmethod
from typing import Any

from .types.async_http_client_types import GetArgs, UploadFileArgs


class AsyncHTTPClientContract(ABC):
    """
    A contract for asynchronous HTTP client implementations.

    This abstract base class defines the required methods for performing
    HTTP GET requests and uploading files asynchronously.

    Methods.
    -------
    get(args: GetArgs) -> dict[str, Any]
        Perform an HTTP GET request with the specified arguments.

    upload_file(args: UploadFileArgs) -> dict[str, Any]
        Upload a file using the specified arguments.
    """

    @abstractmethod
    async def get(self, args: GetArgs) -> dict[str, Any]:
        pass

    @abstractmethod
    async def upload_file(self, args: UploadFileArgs) -> dict[str, Any]:
        pass



class ThiesUploadEmptyError(Exception):
    """Raised when no files are found to upload to the server."""

    def __str__(self):
        return "No files were found to upload."


class FetchCloudFileNamesError(Exception):
    """Raised when there is an error fetching file names from the RCER cloud."""

    def __str__(self):
        return "An error occurred while retrieving file names from the RCER cloud"


class FetchThiesFileContentError(Exception):
    """Raised when there is an error fetching the content of a Thies file."""

    def __str__(self):
        return "An error occurred while retrieving the content of a Thies file"



from http import HTTPStatus

from rcer_iot_client_pkg.general_types.error_types.api.update_thies_data_error_types import (
    FetchCloudFileNamesError,
    ThiesUploadEmptyError,
)
from rcer_iot_client_pkg.general_types.error_types.common.common_types import (
    FtpClientError,
    HttpClientError,
)
from rcer_iot_client_pkg.services.epii.controllers.types.update_thies_data_types import (
    UpdateThiesDataControllerInput,
    UpdateThiesDataControllerOutput,
)
from rcer_iot_client_pkg.services.epii.use_cases.types import (
    UpdateThiesDataUseCaseInput,
)
from rcer_iot_client_pkg.services.epii.use_cases.update_thies_data import (
    UpdateThiesDataUseCase,
)


class UpdateThiesDataController:
    def __init__(self, input: UpdateThiesDataControllerInput):
        self.use_case = UpdateThiesDataUseCase(
            UpdateThiesDataUseCaseInput(**input.__dict__)
        )

    async def execute(self) -> UpdateThiesDataControllerOutput:
        try:
            data = await self.use_case.execute()
            return UpdateThiesDataControllerOutput(
                message="THIES was synced successfully",
                status=HTTPStatus.OK.value,
                metadata={"data": data},
            )
        except (AttributeError, NameError, ValueError) as error:
            return UpdateThiesDataControllerOutput(
                message="An unexpected error occurred during use case initialization.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )
        except FtpClientError as error:
            return UpdateThiesDataControllerOutput(
                message="Ftp Client initialization fails.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except HttpClientError as error:
            return UpdateThiesDataControllerOutput(
                message="Http Client initialization fails.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except FetchCloudFileNamesError as error:
            return UpdateThiesDataControllerOutput(
                message="An error occurred while retrieving file names from the RCER cloud",
                status=HTTPStatus.INTERNAL_SERVER_ERROR.value,
                metadata={"error": error.__str__()},
            )
        except ThiesUploadEmptyError as error:
            return UpdateThiesDataControllerOutput(
                message="No files were found to upload.",
                status=HTTPStatus.NO_CONTENT.value,
                metadata={"error": error.__str__()},
            )



import os

from dotenv import load_dotenv

import rcer_iot_client_pkg.services.epii.use_cases.constants as c
from src.rcer_iot_client_pkg.general_types.error_types.api.update_thies_data_error_types import (
    FetchCloudFileNamesError,
    FetchThiesFileContentError,
    ThiesUploadEmptyError,
)
from rcer_iot_client_pkg.general_types.error_types.common import (
    EmptyDataError,
    FtpClientError,
    HttpClientError,
)
from rcer_iot_client_pkg.libs.async_http_client import (
    AsyncHTTPClient,
    AsyncHttpClientInitArgs,
    GetArgs,
)
from rcer_iot_client_pkg.libs.ftp_client import (
    FTPClient,
    FtpClientInitArgs,
    ListFilesArgs,
    ReadFileArgs,
)
from rcer_iot_client_pkg.services.epii.use_cases.types import (
    UpdateThiesDataUseCaseInput,
)
from rcer_iot_client_pkg.services.epii.utils import (
    generate_file_content,
)

load_dotenv()


class UpdateThiesDataUseCase:
    def __init__(self, input: UpdateThiesDataUseCaseInput):
        self.ftp_port = input.ftp_port
        self.ftp_host = input.ftp_host
        self.ftp_password = input.ftp_password
        self.ftp_user = input.ftp_user
        self.sharepoint_client = self._initialize_sharepoint_client()
        self.thies_ftp_client = self._initialize_thies_ftp_client()
        self.uploading = set()

    def _initialize_sharepoint_client(self) -> AsyncHTTPClient:
        """Initialize the HTTP client."""
        try:
            return AsyncHTTPClient(
                AsyncHttpClientInitArgs(
                    client_name="aiohttp_client",
                    access_token="temporal-token",
                    base_url="https://graph.microsoft.com/v1.0/",
                )
            )
        except ConnectionError as error:
            raise HttpClientError(error)

    def _initialize_thies_ftp_client(self) -> FTPClient:
        """Initialize the FTP client."""
        try:
            return FTPClient(
                FtpClientInitArgs(
                    client_name="aioftp_client",
                    host=self.ftp_host,
                    user=self.ftp_user,
                    password=self.ftp_password,
                    port=self.ftp_port,
                )
            )
        except RuntimeError as error:
            raise FtpClientError(error)

    async def fetch_cloud_file_names(self, folder_name: str) -> set[str]:
        """Fetch file names from the RCER cloud."""
        try:
            cloud_files = set()
            async with self.sharepoint_client:
                for file_type in c.FILE_TYPES:
                    destination_path = f"Onedrive_UC/noveno-semestre/IPRE-RCER/{folder_name}/{file_type}"
                    endpoint = f"drives/{c.DRIVE_ID}/root:/{destination_path}:/children"
                    response = await self.sharepoint_client.get(
                        GetArgs(endpoint=endpoint)
                    )
                    cloud_files.update(
                        {f"{file_type}_{item['name']}" for item in response["value"]}
                    )
            return cloud_files
        except ConnectionError as error:
            raise FetchCloudFileNamesError(error)

    async def fetch_thies_file_names(self) -> set[str]:
        """Fetch file names from the THIES FTP server."""
        try:
            avg_files = await self.thies_ftp_client.list_files(
                ListFilesArgs(path=c.PATH_AVG_FILES)
            )
            ext_files = await self.thies_ftp_client.list_files(
                ListFilesArgs(path=c.PATH_EXT_FILES)
            )
            return {f"AVG_{name}" for name in avg_files} | {
                f"EXT_{name}" for name in ext_files
            }
        except ConnectionError:
            raise ThiesUploadEmptyError

    async def fetch_thies_file_content(self) -> dict[str, bytes]:
        """Fetch the content of files from the THIES FTP server."""
        content_files = {}
        for file in self.uploading:
            try:
                origin, filename = file.split("_", 1)
                file_path = (
                    f"{c.PATH_AVG_FILES}/{filename}"
                    if origin == "AVG"
                    else f"{c.PATH_EXT_FILES}/{filename}"
                )
                content = await self.thies_ftp_client.read_file(ReadFileArgs(file_path))
                content_files[filename] = content
            except ConnectionError as error:
                raise FetchThiesFileContentError(error)
        return content_files

    async def execute(self) -> dict:
        """Synchronize data from the THIES Center to the cloud."""
        try: 
            thies_files = await self.fetch_thies_file_names()
        except RuntimeError as error:
            raise FtpClientError(error)
        
        cloud_files = await self.fetch_cloud_file_names(folder_name="thies")
        self.uploading = thies_files - cloud_files
        if not self.uploading:
            raise EmptyDataError

        thies_file_contents = await self.fetch_thies_file_content()
        data = generate_file_content(thies_file_contents)
        return data



from aioftp import Client

from rcer_iot_client_pkg.libs.ftp_client.ftp_client_contract import (
    FTPClientContract,
)
from rcer_iot_client_pkg.libs.ftp_client.types.ftp_client_types import (
    FtpClientInitArgs,
    ListFilesArgs,
    ReadFileArgs,
)


class AioFTPClient(FTPClientContract):
    def __init__(self, args: FtpClientInitArgs) -> None:
        self.host = args.host
        self.port = args.port
        self.password = args.password
        self.user = args.user
        self.client = Client()

    async def _async_start(self) -> None:
        try: 
            await self.client.connect(host=self.host, port=self.port)
            await self.client.login(user=self.user, password=self.password)
        except Exception:
            raise RuntimeError("Unexpected error occurred while trying to connect to the FTP server")

    async def list_files(self, args: ListFilesArgs) -> list[str]:
        await self._async_start()
        return [
            path.name async for path, _ in self.client.list(args.path, recursive=False)
        ]

    async def read_file(self, args: ReadFileArgs) -> bytes:
        await self._async_start()
        async with self.client.download_stream(args.file_path) as stream:
            return await stream.read()
